@@ -28,6 +28,9 @@ import copy
 from collections import namedtuple
 from warnings import warn
 from abc import abstractmethod
+import inspect
+from collections import defaultdict
+import re
 
 import numpy as np
 from sklearn.base import clone
@@ -190,8 +193,9 @@ def _crossfit(model, folds, *args, **kwargs):
     return nuisances, model_list, np.sort(fitted_inds.astype(int)), (scores if calculate_scores else None)
 
 
-CachedValues = namedtuple('_CachedValues', ['nuisances',
-                                            'Y', 'T', 'X', 'W', 'Z', 'sample_weight', 'sample_var', 'groups'])
+CachedValues = namedtuple('CachedValues', ['nuisances',
+                                           'Y', 'T', 'X', 'W', 'Z', 'sample_weight', 'freq_weight',
+                                           'sample_var', 'groups'])
 
 
 class _OrthoLearner(TreatmentExpansionMixin, LinearCateEstimator):
@@ -232,7 +236,7 @@ class _OrthoLearner(TreatmentExpansionMixin, LinearCateEstimator):
 
         .. math ::
             \\mathbb{E}_n[\\ell(V; \\theta(X), \\hat{h}(V))]\
-            = \\frac{1}{n} \\sum_{i=1}^n \\sum_i \\ell(V_i; \\theta(X_i), \\hat{U}_i)
+            = \\frac{1}{n} \\sum_{i=1}^n \\ell(V_i; \\theta(X_i), \\hat{U}_i)
 
         The method is a bit more general in that the final step does not need to be a loss minimization step.
         The class takes as input a model for fitting an estimate of the nuisance h given a set of samples
@@ -408,26 +412,24 @@ class _OrthoLearner(TreatmentExpansionMixin, LinearCateEstimator):
 
     Attributes
     ----------
-    models_nuisance_: list of objects of type(model_nuisance)
-        A list of instances of the model_nuisance object. Each element corresponds to a crossfitting
-        fold and is the model instance that was fitted for that training fold. If `mc_iters` is > 1,
-        then the fitted models from the last monte carlo iteration are being stored and returned.
-        TODO. Enable returning all fitted nuisance models from all monte carlo iterations.
+    models_nuisance_: nested list of objects of type(model_nuisance)
+        A nested list of instances of the model_nuisance object. The number of sublist equals to the
+        number of monte carlo iterations. Each element in the sublist corresponds to a crossfitting
+        fold and is the model instance that was fitted for that training fold.
     ortho_learner_model_final_: object of type(model_final)
         An instance of the model_final object that was fitted after calling fit.
     score_ : float or array of floats
         If the model_final has a score method, then `score_` contains the outcome of the final model
         score when evaluated on the fitted nuisances from the first stage. Represents goodness of fit,
         of the final CATE model.
-    nuisance_scores_ : tuple of lists of floats or None
+    nuisance_scores_ : tuple of nested lists of floats or None
         The out-of-sample scores from training each nuisance model
     """
 
     def __init__(self, *,
                  discrete_treatment, discrete_instrument, categories, cv, random_state,
-                 n_splits='raise', mc_iters=None, mc_agg='mean'):
+                 mc_iters=None, mc_agg='mean'):
         self.cv = cv
-        self.n_splits = n_splits
         self.discrete_treatment = discrete_treatment
         self.discrete_instrument = discrete_instrument
         self.random_state = random_state
@@ -447,9 +449,9 @@ class _OrthoLearner(TreatmentExpansionMixin, LinearCateEstimator):
             `fit` and `predict` methods that both have signatures::
 
                 model_nuisance.fit(Y, T, X=X, W=W, Z=Z,
-                                sample_weight=sample_weight, sample_var=sample_var)
+                                sample_weight=sample_weight)
                 model_nuisance.predict(Y, T, X=X, W=W, Z=Z,
-                                    sample_weight=sample_weight, sample_var=sample_var)
+                                    sample_weight=sample_weight)
 
             In fact we allow for the model method signatures to skip any of the keyword arguments
             as long as the class is always called with the omitted keyword argument set to ``None``.
@@ -472,7 +474,7 @@ class _OrthoLearner(TreatmentExpansionMixin, LinearCateEstimator):
             Must implement `fit` and `predict` methods that must have signatures::
 
                 model_final.fit(Y, T, X=X, W=W, Z=Z, nuisances=nuisances,
-                                sample_weight=sample_weight, sample_var=sample_var)
+                                sample_weight=sample_weight, freq_weight=freq_weight, sample_var=sample_var)
                 model_final.predict(X=X)
 
             Predict, should just take the features X and return the constant marginal effect. In fact we allow
@@ -514,7 +516,7 @@ class _OrthoLearner(TreatmentExpansionMixin, LinearCateEstimator):
         return var[inds] if var is not None else None
 
     def _strata(self, Y, T, X=None, W=None, Z=None,
-                sample_weight=None, sample_var=None, groups=None,
+                sample_weight=None, freq_weight=None, sample_var=None, groups=None,
                 cache_values=False, only_final=False, check_input=True):
         if self.discrete_instrument:
             Z = LabelEncoder().fit_transform(np.ravel(Z))
@@ -544,7 +546,7 @@ class _OrthoLearner(TreatmentExpansionMixin, LinearCateEstimator):
     @_deprecate_positional("X, W, and Z should be passed by keyword only. In a future release "
                            "we will disallow passing X, W, and Z by position.", ['X', 'W', 'Z'])
     @BaseCateEstimator._wrap_fit
-    def fit(self, Y, T, X=None, W=None, Z=None, *, sample_weight=None, sample_var=None, groups=None,
+    def fit(self, Y, T, X=None, W=None, Z=None, *, sample_weight=None, freq_weight=None, sample_var=None, groups=None,
             cache_values=False, inference=None, only_final=False, check_input=True):
         """
         Estimate the counterfactual model from data, i.e. estimates function :math:`\\theta(\\cdot)`.
@@ -561,10 +563,15 @@ class _OrthoLearner(TreatmentExpansionMixin, LinearCateEstimator):
             Controls for each sample
         Z: optional (n, d_z) matrix or None (Default=None)
             Instruments for each sample
-        sample_weight: optional (n,) vector or None (Default=None)
-            Weights for each samples
-        sample_var: optional (n,) vector or None (Default=None)
-            Sample variance for each sample
+        sample_weight : (n,) array like, default None
+            Individual weights for each sample. If None, it assumes equal weight.
+        freq_weight: (n, ) array like of integers, default None
+            Weight for the observation. Observation i is treated as the mean
+            outcome of freq_weight[i] independent observations.
+            When ``sample_var`` is not None, this should be provided.
+        sample_var : {(n,), (n, d_y)} nd array like, default None
+            Variance of the outcome(s) of the original freq_weight[i] observations that were used to
+            compute the mean outcome represented by observation i.
         groups: (n,) vector, optional
             All rows corresponding to the same group will be kept together during splitting.
             If groups is not None, the cv argument passed to this class's initializer
@@ -588,10 +595,12 @@ class _OrthoLearner(TreatmentExpansionMixin, LinearCateEstimator):
         self : object
         """
         self._random_state = check_random_state(self.random_state)
+        assert (freq_weight is None) == (
+            sample_var is None), "Sample variances and frequency weights must be provided together!"
         if check_input:
-            Y, T, X, W, Z, sample_weight, sample_var, groups = check_input_arrays(
-                Y, T, X, W, Z, sample_weight, sample_var, groups)
-            self._check_input_dims(Y, T, X, W, Z, sample_weight, sample_var, groups)
+            Y, T, X, W, Z, sample_weight, freq_weight, sample_var, groups = check_input_arrays(
+                Y, T, X, W, Z, sample_weight, freq_weight, sample_var, groups)
+            self._check_input_dims(Y, T, X, W, Z, sample_weight, freq_weight, sample_var, groups)
 
         if not only_final:
 
@@ -613,10 +622,30 @@ class _OrthoLearner(TreatmentExpansionMixin, LinearCateEstimator):
 
             all_nuisances = []
             fitted_inds = None
+            if sample_weight is None:
+                if freq_weight is not None:
+                    sample_weight_nuisances = freq_weight
+                else:
+                    sample_weight_nuisances = None
+            else:
+                if freq_weight is not None:
+                    sample_weight_nuisances = freq_weight * sample_weight
+                else:
+                    sample_weight_nuisances = sample_weight
 
-            for _ in range(self.mc_iters or 1):
-                nuisances, new_inds = self._fit_nuisances(Y, T, X, W, Z, sample_weight=sample_weight, groups=groups)
+            self._models_nuisance = []
+            for idx in range(self.mc_iters or 1):
+                nuisances, fitted_models, new_inds, scores = self._fit_nuisances(
+                    Y, T, X, W, Z, sample_weight=sample_weight_nuisances, groups=groups)
                 all_nuisances.append(nuisances)
+                self._models_nuisance.append(fitted_models)
+                if scores is None:
+                    self.nuisance_scores_ = None
+                else:
+                    if idx == 0:
+                        self.nuisance_scores_ = tuple([] for _ in scores)
+                    for ind, score in enumerate(scores):
+                        self.nuisance_scores_[ind].append(score)
                 if fitted_inds is None:
                     fitted_inds = new_inds
                 elif not np.array_equal(fitted_inds, new_inds):
@@ -633,12 +662,14 @@ class _OrthoLearner(TreatmentExpansionMixin, LinearCateEstimator):
                     raise ValueError(
                         "Parameter `mc_agg` must be one of {'mean', 'median'}. Got {}".format(self.mc_agg))
 
-            Y, T, X, W, Z, sample_weight, sample_var = (self._subinds_check_none(arr, fitted_inds)
-                                                        for arr in (Y, T, X, W, Z, sample_weight, sample_var))
+            Y, T, X, W, Z, sample_weight, freq_weight, sample_var = (self._subinds_check_none(arr, fitted_inds)
+                                                                     for arr in (Y, T, X, W, Z, sample_weight,
+                                                                                 freq_weight, sample_var))
             nuisances = tuple([self._subinds_check_none(nuis, fitted_inds) for nuis in nuisances])
             self._cached_values = CachedValues(nuisances=nuisances,
                                                Y=Y, T=T, X=X, W=W, Z=Z,
                                                sample_weight=sample_weight,
+                                               freq_weight=freq_weight,
                                                sample_var=sample_var,
                                                groups=groups) if cache_values else None
         else:
@@ -648,9 +679,12 @@ class _OrthoLearner(TreatmentExpansionMixin, LinearCateEstimator):
             if self.transformer is not None:
                 self._d_t = (len(self.transformer.categories_[0]) - 1,)
 
-        self._fit_final(Y=Y, T=T, X=X, W=W, Z=Z,
+        self._fit_final(Y=Y,
+                        T=self.transformer.transform(T.reshape((-1, 1))) if self.transformer is not None else T,
+                        X=X, W=W, Z=Z,
                         nuisances=nuisances,
                         sample_weight=sample_weight,
+                        freq_weight=freq_weight,
                         sample_var=sample_var)
 
         return self
@@ -685,7 +719,7 @@ class _OrthoLearner(TreatmentExpansionMixin, LinearCateEstimator):
         cached = self._cached_values
         kwargs = filter_none_kwargs(
             Y=cached.Y, T=cached.T, X=cached.X, W=cached.W, Z=cached.Z,
-            sample_weight=cached.sample_weight, sample_var=cached.sample_var,
+            sample_weight=cached.sample_weight, freq_weight=cached.freq_weight, sample_var=cached.sample_var,
             groups=cached.groups,
         )
         _OrthoLearner.fit(self, **kwargs,
@@ -733,21 +767,21 @@ class _OrthoLearner(TreatmentExpansionMixin, LinearCateEstimator):
         nuisances, fitted_models, fitted_inds, scores = _crossfit(self._ortho_learner_model_nuisance, folds,
                                                                   Y, T, X=X, W=W, Z=Z,
                                                                   sample_weight=sample_weight, groups=groups)
-        self._models_nuisance = fitted_models
-        self.nuisance_scores_ = scores
-        return nuisances, fitted_inds
+        return nuisances, fitted_models, fitted_inds, scores
 
-    def _fit_final(self, Y, T, X=None, W=None, Z=None, nuisances=None, sample_weight=None, sample_var=None):
+    def _fit_final(self, Y, T, X=None, W=None, Z=None, nuisances=None, sample_weight=None,
+                   freq_weight=None, sample_var=None):
         self._ortho_learner_model_final.fit(Y, T, **filter_none_kwargs(X=X, W=W, Z=Z,
                                                                        nuisances=nuisances,
                                                                        sample_weight=sample_weight,
+                                                                       freq_weight=freq_weight,
                                                                        sample_var=sample_var))
         self.score_ = None
         if hasattr(self._ortho_learner_model_final, 'score'):
             self.score_ = self._ortho_learner_model_final.score(Y, T, **filter_none_kwargs(X=X, W=W, Z=Z,
                                                                                            nuisances=nuisances,
-                                                                                           sample_weight=sample_weight,
-                                                                                           sample_var=sample_var))
+                                                                                           sample_weight=sample_weight)
+                                                                )
 
     def const_marginal_effect(self, X=None):
         X, = check_input_arrays(X)
@@ -786,9 +820,9 @@ class _OrthoLearner(TreatmentExpansionMixin, LinearCateEstimator):
         """
         Score the fitted CATE model on a new data set. Generates nuisance parameters
         for the new data set based on the fitted nuisance models created at fit time.
-        It uses the mean prediction of the models fitted by the different crossfit folds.
-        Then calls the score function of the model_final and returns the calculated score.
-        The model_final model must have a score method.
+        It uses the mean prediction of the models fitted by the different crossfit folds
+        under different iterations. Then calls the score function of the model_final and
+        returns the calculated score. The model_final model must have a score method.
 
         If model_final does not have a score method, then it raises an :exc:`.AttributeError`
 
@@ -821,17 +855,22 @@ class _OrthoLearner(TreatmentExpansionMixin, LinearCateEstimator):
         X, T = self._expand_treatments(X, T)
         if self.z_transformer is not None:
             Z = self.z_transformer.transform(reshape(Z, (-1, 1)))
-        n_splits = len(self._models_nuisance)
-        for idx, mdl in enumerate(self._models_nuisance):
-            nuisance_temp = mdl.predict(Y, T, **filter_none_kwargs(X=X, W=W, Z=Z))
-            if not isinstance(nuisance_temp, tuple):
-                nuisance_temp = (nuisance_temp,)
+        n_iters = len(self._models_nuisance)
+        n_splits = len(self._models_nuisance[0])
 
-            if idx == 0:
-                nuisances = [np.zeros((n_splits,) + nuis.shape) for nuis in nuisance_temp]
+        # for each mc iteration
+        for i, models_nuisances in enumerate(self._models_nuisance):
+            # for each model under cross fit setting
+            for j, mdl in enumerate(models_nuisances):
+                nuisance_temp = mdl.predict(Y, T, **filter_none_kwargs(X=X, W=W, Z=Z))
+                if not isinstance(nuisance_temp, tuple):
+                    nuisance_temp = (nuisance_temp,)
 
-            for it, nuis in enumerate(nuisance_temp):
-                nuisances[it][idx] = nuis
+                if i == 0 and j == 0:
+                    nuisances = [np.zeros((n_iters * n_splits,) + nuis.shape) for nuis in nuisance_temp]
+
+                for it, nuis in enumerate(nuisance_temp):
+                    nuisances[it][i * n_iters + j] = nuis
 
         for it in range(len(nuisances)):
             nuisances[it] = np.mean(nuisances[it], axis=0)
@@ -850,18 +889,3 @@ class _OrthoLearner(TreatmentExpansionMixin, LinearCateEstimator):
         if not hasattr(self, '_models_nuisance'):
             raise AttributeError("Model is not fitted!")
         return self._models_nuisance
-
-    #######################################################
-    # These should be removed once `n_splits` is deprecated
-    #######################################################
-
-    @property
-    def n_splits(self):
-        return self.cv
-
-    @n_splits.setter
-    def n_splits(self, value):
-        if value != 'raise':
-            warn("Parameter `n_splits` has been deprecated and will be removed in the next version. "
-                 "Use parameter `cv` instead.")
-            self.cv = value
